/* 
 * WorldWide File System
 * Copyright (c) 1992,1993 Youki Kadobayashi
 * Copyright (c) 1992,1993 Osaka University
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that the following conditions
 * are met:
 * 1. Both the copyright notice and this permission notice appear in
 *    all copies of the software, derivative works or modified versions,
 *    and any portions thereof, and that both notices appear in
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Osaka University
 *      and its contributors.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPER ``AS IS'' AND OSAKA
 * UNIVERSITY DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Osaka University requests users of this software to return to
 *
 *  Youki Kadobayashi
 *  Department of Information and Computer Sciences
 *  Osaka University, Toyonaka 560, Osaka, Japan
 *
 * any improvements or extensions that they make and grant Osaka
 * University the rights to redistribute these changes.
 */
/* cmfs: cache manager file system.
 * Perform access to cached information. Call delegate function xyz_miss()
 * when cache missed.
 */
static char *AtFSid = "$Header: cmfs.c[109.0] Wed Nov 24 03:47:03 1993 youki-k@is.aist-nara.ac.jp saved $";

#include "wfs.h"
#include "util.h"
#include "global.h"

/*
 * filehandle internals

 * parent				child
 * ---------				--------
 * parent_vol				---
 * parent_dir				---
 * child_vol			==	parent_vol
 * child_dir			==	dir_id
 * file_findid(...)->fsid	==	child_vol
 * file_findid(...)->id		==	file_id

 * parent				child "."
 * ---------				--------
 * parent_vol			==	parent_vol
 * parent_dir			==	parent_dir
 * child_vol			==	child_vol
 * child_dir			==	child_dir

 * parent				child ".."
 * ---------				--------
 * file_findname(, "..")->fsid	==	parent_vol
 * file_findname(, "..")->id	==	parent_dir
 * parent_vol			==	child_vol
 * parent_dir			==	child_dir
 * child_vol				---
 * child_dir				---
 *
 * Note that dir_id and file_id is aliased as
 * parent_dir and child_dir, respectively.
 */
void
cmfs_getattr(c)
wf_thrd *c;
/* in: rqstp, volp, fhp, reply */
/* out: dirp, filep */
{
	static wf_file file;

	c->dirp = dir_findcache(c->parent_volp, c->fhp->dir_id);
	if (c->dirp == (wf_dir *)0) {
#if 0
		(*c->reply)(c, WF_ERR_NOENT);
		return;
#else
		/* fake that directory for nonexistent directory */
		static wf_dir dir;
		dir.id = c->fhp->dir_id;
		c->dirp = &dir;
		c->filep = (wf_file *)0;
#endif
	} else {
		c->filep = file_findid(c->dirp,
				       c->fhp->child_vol, c->fhp->file_id);
	}
	if (!c->filep) {
		/* fake that file for nonexistent directory */
		attr_initdir(&file.fattr);
		file.fattr.fsid = c->fhp->child_vol;
		file.id = file.fattr.fileid = c->fhp->file_id;
		c->filep = &file;
	}
	if (c->filep->flag & WF_FILE_MUTABLE) {
		/* Delegate gets notification of access, but not
		 * required to reflect current state right now.
		 * Getattr operation should be "cheap" however.
		 * Attribute caching mechanism is deferred to each
		 * underlying filesystem for now.
		 */
		proto_getattr_miss(c);
	}
	(*c->reply)(c, WF_REP_OK);
}

void
cmfs_lookup(c)
wf_thrd *c;
/* in: rqstp, volp, fhp, fname, reply */
/* out: dotdotp, resfhp, filep */
{
	wf_fh		*resfhp;	/* reply filehandle */

	/* stats */
	c->child_volp->stats.n_request_lookup++;

	if (c->fname[0] == '.') {
		wf_dir *dotdotp;

		dotdotp = dir_findcache(c->parent_volp,
					c->fhp->parent_dir);
		if (dotdotp == NULL || dotdotp->flag & WF_DIR_OBSOLETE) {
			(*c->reply)(c, WF_ERR_STALE);
		} else if (strcmp(c->fname, "..") == 0) {
			/*
			 * dotdot
			 */
			c->resfhp = resfhp = CLONE(c->fhp);
			c->filep = file_findname(dotdotp, "..");
			assert(c->filep != NULL);
			resfhp->parent_vol = FSID(c->filep);
			resfhp->parent_dir = FID(c->filep);

			c->filep = file_findname(dotdotp, ".");
			assert(c->filep != NULL);
			resfhp->child_vol = c->fhp->parent_vol;
			resfhp->child_dir = c->fhp->parent_dir;
			(*c->reply)(c, WF_REP_OK);
			FREE(resfhp);
			return;
		} else if (strcmp(c->fname, ".") == 0) {
			/*
			 * dot
			 */
			c->filep = file_findid(dotdotp,
					       c->fhp->child_vol, c->fhp->child_dir);
			c->resfhp = c->fhp;
			(*c->reply)(c, WF_REP_OK);
			return;
		}
	}

	/*
	 * get dotdotp
	 */
	c->dotdotp = dir_findcache(c->child_volp, c->fhp->child_dir);
	if (c->dotdotp == (wf_dir *)0 || c->dotdotp->flag & WF_DIR_OBSOLETE) {
		(*c->reply)(c, WF_ERR_STALE);
		return;
	}

	/*
	 * find file in dotdotp
	 */
	c->filep = file_findname(c->dotdotp, c->fname);
	if (c->filep == (wf_file *)0) {
		c->child_volp->stats.n_request_lookup_miss++;
		proto_lookup_miss(c);
		return;
	}

	c->resfhp = resfhp = CLONE(c->fhp);
	resfhp->parent_vol = c->fhp->child_vol;
	resfhp->parent_dir = c->fhp->child_dir;
	resfhp->child_vol = FSID(c->filep);
	resfhp->child_dir = FID(c->filep);
	(*c->reply)(c, WF_REP_OK);
	FREE(resfhp);
}

#ifdef TRACE
static wf_trace *
cmfs_readdir_trace(c)
wf_thrd *c;
{
	wf_trace *tracep;

	tracep = ALLOC(wf_trace);
	tracep->time = gettime();
	tracep->uid = c->uid;
	tracep->client = c->client;
	tracep->volume = c->fhp->child_vol;
	tracep->dir = c->fhp->child_dir;
	return tracep;
}
#endif

void
cmfs_readdir(c)
wf_thrd *c;
/* in: rqstp, volp, fhp, cookie, count, reply */
/* out: dirp, dotdotp, olddirp, filep */
{
	/*
	 * assumption: consecutive queries will come in before
	 * asked directory gets kicked out of memory
	 */
#ifdef TRACE
	wf_trace *tracep;
#endif

	/* stats */
	c->child_volp->stats.n_request_readdir++;

	/* find file in dotdotp */
	c->dotdotp = dir_findcache(c->parent_volp, c->fhp->parent_dir);
	if (c->dotdotp == NULL) {
		(*c->reply)(c, WF_ERR_NOENT);
		return;
	}
	c->filep = file_findid(c->dotdotp,
			       c->fhp->child_vol, c->fhp->child_dir);
	if (c->filep == NULL) {
		/* file handle corrupt? */
		(*c->reply)(c, WF_ERR_NOENT);
		return;
	}

	/*
	 * check if directory transfer is in progress
	 */
	if (c->filep->thrdp) {
		(*c->reply)(c, WF_ERR_ALREADY);
		return;
	}

	/*
	 * get dirp
	 */
	c->dirp = dir_findcache(c->child_volp, c->fhp->child_dir);

	/*
	 * cache hit
	 */
	if (c->dirp && !(c->dirp->flag & WF_DIR_OBSOLETE)) {
#ifdef TRACE
		/*
		 * trace cache hit
		 */
		unsigned long cookie;
		bcopy(c->cookie, &cookie, sizeof(nfscookie));
		if (cookie == 0) {
			tracep = cmfs_readdir_trace(c);
			tracep->op = WF_OP_READDIR;
			tracep->hit = WF_CACHE_HIT;
			trace_event(c, tracep);
		}
#endif
		(*c->reply)(c, WF_REP_OK);
		return;
	}

	/*
	 * cache missed
	 */
	c->child_volp->stats.n_request_readdir_miss++;
	c->olddirp = c->dirp;
#ifdef TRACE
	/*
	 * trace cache miss
	 */
	tracep = cmfs_readdir_trace(c);
	tracep->op = WF_OP_READDIR;
	tracep->hit = WF_CACHE_MISS;
	trace_event(c, tracep);
#endif
	proto_readdir_miss(c);
}

void
cmfs_readlink(c)
wf_thrd *c;
/* in: rqstp, volp, fhp, reply */
/* out: dirp, filep */
{
	c->dirp = dir_findcache(c->parent_volp, c->fhp->dir_id);
	if (c->dirp == (wf_dir *)0 || c->dirp->flag & WF_DIR_OBSOLETE) {
		(*c->reply)(c, WF_ERR_STALE);
		return;
	}
	c->filep = file_findid(c->dirp, c->fhp->child_vol, c->fhp->file_id);
	assert(c->filep != (wf_file *)0);
	if (c->filep->fattr.type != NFLNK) {
		(*c->reply)(c, WF_ERR_NOENT);
		return;
	}
	(*c->reply)(c, WF_REP_OK);
}

#ifdef TRACE
static wf_trace *
cmfs_read_trace(c)
wf_thrd *c;
{
	wf_trace *tracep;

	tracep = ALLOC(wf_trace);
	tracep->time = gettime();
	tracep->uid = c->uid;
	tracep->client = c->client;
	tracep->volume = c->fhp->child_vol;
	tracep->dir = c->fhp->dir_id;
	tracep->file = c->fhp->file_id;
	tracep->arg = (long) c->filep->fattr.size;
	return tracep;
}
#endif

void
cmfs_read(c)
wf_thrd *c;
/* in: rqstp, volp, fhp, offset, count, totalcount, reply */
/* out: dirp, filep */
{
#ifdef TRACE
	wf_trace *tracep;
#endif
	/* stats */
	c->child_volp->stats.n_request_read++;

	/*
	 * get dirp
	 */
	c->dirp = dir_findcache(c->parent_volp, c->fhp->dir_id);
	if (c->dirp == (wf_dir *)0 || c->dirp->flag & WF_DIR_OBSOLETE) {
		(*c->reply)(c, WF_ERR_STALE);
		return;
	}

	/*
	 * find file in dirp
	 */
	c->filep = file_findid(c->dirp, c->fhp->child_vol, c->fhp->file_id);
	assert(c->filep != (wf_file *)0);

	/*
	 * check if file transfer is in progress
	 */
	if (c->filep->thrdp) {
		(*c->reply)(c, WF_ERR_ALREADY);
		return;
	}

	/*
	 * cache hit
	 */
	if (file_findcache(c->child_volp, c->dirp, c->filep) == 0) {
#ifdef TRACE
		/*
		 * trace cache hit
		 */
		if (c->offset == 0) {
			tracep = cmfs_read_trace(c);
			tracep->op = WF_OP_READ;
			tracep->hit = WF_CACHE_HIT;
			trace_event(c, tracep);
		}
#endif
		(*c->reply)(c, WF_REP_OK);
		return;
	}

	/*
	 * cache missed
	 */
	c->child_volp->stats.n_request_read_miss++;
#ifdef TRACE
	/*
	 * trace cache missed
	 */
	tracep = cmfs_read_trace(c);
	tracep->op = WF_OP_READ;
	tracep->hit = WF_CACHE_MISS;
	trace_event(c, tracep);
#endif
	proto_read_miss(c);
}
