#!/usr/local/bin/perl
#
# gtr.pl -- GTR library.  GTR stands for geographical / topological region.
# Author: Youki Kadobayashi <youki@wide.ad.jp>
# This file is part of WWFS.
#
package gtr;

# configuration parameters:
# $gtr'dir	gtr directory; "/etc" by default.
# $gtr'conf	configuration file. "/etc/gtr.conf" by default.
# $gtr'txt	address-to-GTR mapping table. "/etc/gtr.txt" by default.

$conf = $dir ? "$dir/gtr.conf" : "/etc/gtr.conf";
$txt = $dir ? "$dir/gtr.txt" : "/etc/gtr.txt";

#
# @sorted_hostnames = gtr'sort(@hostnames);
# get a list of hostnames and sort them according to /etc/gtr.*
#
sub sort {
    local (@hosts) = @_;
    local (@hostlist, @sorted, @gtrs, @addrs, @list);
    local ($name, $aliases, $addrtype, $length);
    local ($host, $ip, $rating, $entry);

    foreach $host (@hosts) {
	($name,$aliases,$addrtype,$length,@addrs) = gethostbyname($host);
	@gtrs = ();
	# each host may have multiple interfaces...
	foreach $addr (@addrs) {
	    # For now, assume IPv4 address and bypass checking.
	    # need to adapt to IPv7 in near future.
	    $ip = join('.', unpack('C4', $addr));
	    push(@gtrs, &gtr'byaddr($ip));
	}
	$rating = &gtr'select(@gtrs);
	$list[$rating] .= "$host:";
    }

    foreach $entry (@list) {
	chop $entry;		# trailing colon
	@hostlist = split(/:/, $entry);
	if ($#hostlist >= 0) {
	    push(@sorted, @hostlist);
	    #  print join(' ', @hostlist), "\n";
	}
    }
    return @sorted;
}

sub ip2long {
    local ($str) = @_;
    local (@a);
    # ip address to long conversion.

    @a = split(/\./, $str);
    push(@a, (0, 0, 0));	# permit "133.1" style of abbreviation
    return $a[0] * 0x01000000 + $a[1] * 0x010000 + $a[2] * 0x0100 + $a[3];
}

#
# @gtrs = gtr'byaddr($ip_addr_string);
# get a list of GTR from an IP address.
#
sub byaddr {
    local ($addr_string) = @_;
    local ($addr, @gtrs, @line);
    local ($gtr_addr, $gtr_mask, $gtr_name);

    $addr = &ip2long($addr_string);
    open(GTR, $txt) || return undef;
    while (<GTR>) {
	chop;
	s/#.*$//;		# ignore comments
	next if /^\s*$/;
	@line = split(/\s+/, $_);

	if ($line[0] =~ /^[0-9.]+$/) {
	    $gtr_addr = &ip2long($line[0]);
	} else {
	    warn "$txt line $.: IP address expected.\n";
	    next;
	}

	# for abbreviation & fool-proof purpose.
	if ($#line >= 2) {
	    if ($line[1] =~ /^255/) {
		$gtr_mask = &ip2long($line[1]);
	    } else {
		warn "$txt line $.: address mask expected.\n";
		next;
	    }
	    $gtr_name = $line[2];
	} elsif ($#line == 1) {
	    if ($line[1] =~ /^[0-9.]+$/) {
		# you nuts!
		warn "$txt line $.: GTR expected.\n";
		next;
	    }
	    $gtr_name = $line[1];
	    # assume implicit address mask.
	    ($gtr_addr & 0x80000000) == 0 && do {
		$gtr_mask = &ip2long("255.0.0.0");
	    };
	    ($gtr_addr & 0xc0000000) == 0x80000000 && do {
		$gtr_mask = &ip2long("255.255.0.0");
	    };
	    ($gtr_addr & 0xe0000000) == 0xc0000000 && do {
		$gtr_mask = &ip2long("255.255.255.0");
	    }
	} else {
	    warn "$txt line $.: syntax error.\n";
	}

	# test if this GTR matches specified IP address & mask.
	if (($addr & $gtr_mask) == $gtr_addr) {
	    push(@gtrs, $gtr_name);
	}
    }
    close(GTR);
    return (@gtrs);

# Example of /etc/gtr.txt:
#	133.1.192	255.255.255.0	Osaka-ODINS-JP
#	133.1		255.255.0.0	Osaka-WIDE-JP
#	133.4		255.255.0.0	Japan-WIDE-JP
#	192.1.192	255.255.192.0	Foo-Bar-JP
}

#
# $rating = gtr'select(@gtrs)
# select a best-maching GTR (wrt /etc/gtr.conf) from a list of GTRs.
# return a preference value in integer; smaller is better.
#
sub select {
    local (@gtrs) = @_;
    local ($i, $region, $gtr);

    open(CONF, $conf) || return 1000;
    $i = 0;
    while (<CONF>) {
	chop;
	/region\s+/ && do {
	    $region = $';
	    ++$i;

	    foreach $gtr (@gtrs) {
		if ($gtr =~ /$region/) {
		    close(CONF);
		    return $i;
		}
	    }
	};
	# otherwise ignore lines for now.
    }
    close(CONF);
    return 1000;
}

1;
