#!/usr/local/bin/perl
#
# htget               --- recursively get HTMLs starting at a given URL
#
# Given a starting URL, htget will recursively retrieve HTML pages.
# It will also generate "htget.log" -- a table of titles and hosts.
# Tries to find as many new hosts as possible, and strictly
# limits the number of pages it will request from any one server.
#
# htget will stop when $maxtotal pages are retrieved (1000!),
# or when the all pages were retrieved,
# or when SIGINT is received (^C).
#
# NB: to get all pages recursively, try:
#       htget -ls <home-page>
#
# Author: Youki Kadobayashi <youki@wide.ad.jp>
# derived from "explore" written by: Oscar Nierstrasz oscar@cui.unige.ch
# This file is part of WWFS.
#
#v = '(v1.0)'; # August 30, 1993
#v = '(v1.1)'; # August 31 -- added triggering of xmosaic
#v = '(v1.2)'; # Sept 1 -- added various options; SIGINT handling
#v = '(v1.3)'; # Oct 21 -- fixed counting of hosts; added -d
#v = '(v1.4)'; # Oct 23 -- fixed sigint to allow <CR> to continue
#                #       -- fixed printing of $hostsig
#v = '(v1.0)'; # Nov 14, 1993 -- htget initial revision
$v = '(v1.1)'; # Dec 8, 1993 -- htget initial revision

require '/etc/csd.pl';
unshift(@INC, "$WWFSDIR/lib");
require 'url.pl';
require 'dirutil.pl';

$usg = 'Usage: htget [<options>] <http-url> <output file>
	<http-url>	-- URL to start with (no default)
	<output file>	-- default is htget.log
	-m <maxpages>	-- max pages to get per site (default 5)
	-t <maxtotal>	-- max total pages to get (default 100)
	-h <maxhosts>	-- max hosts to explore (default unlimited)
	-s <savedir>	-- directory to save pages to (default current)
	-a <altdir>	-- alternate directory to search
	-x <regexp>	-- exclude files matching regexp
	-ls		-- list all pages at starting site (use with care!)
	-i		-- also get embedded images
';
$maxpages = 5;          # max pages to retrieve per site
$maxtotal = 100;        # max pages to retrieve in total
$maxhosts = undef;      # max hosts to visit

$hosts = 1;             # hosts visited (always at least 1)

chop($date = `date +%d.%m.%y`);
$sig = "This page was generated by htget $v on $date.\n";

# A good default starting point:
# $whatsnew = "http://www.ncsa.uiuc.edu/SDG/Software/Mosaic/Docs/whats-new.html";
# $start = "http://www.ncsa.uiuc.edu/SDG/Software/Mosaic/StartingPoints/NetworkStartingPoints.html";

# default log file:
$deflog = "htget.log";

# default save directory:
$savedir = ".";

while ($#ARGV >= $[) {
	$arg = shift @ARGV;

	if ($arg eq "-m") {
		$arg = shift @ARGV;
		if ($arg =~ /^\d+$/){
			print STDERR "maxpages = $arg (was $maxpages)\n";
			$maxpages = $arg;
			next;
		}
		else { die "Bad arg for -m\n$usg"; }
	}

	if ($arg eq "-t") {
		$arg = shift @ARGV;
		if ($arg =~ /^\d+$/){
			print STDERR "maxtotal = $arg (was $maxtotal)\n";
			$maxtotal = $arg;
			next;
		}
		else { die "Bad arg for -t\n$usg"; }
	}

	if ($arg eq "-h") {
		$arg = shift @ARGV;
		if ($arg =~ /^\d+$/){
			print STDERR "maxhosts = $arg (was $maxhosts)\n";
			$maxhosts = $arg;
			next;
		}
		else { die "Bad arg for -h\n$usg"; }
	}

	if ($arg eq "-s") {
		$arg = shift @ARGV;
		if (-d $arg) {
			print STDERR "directory = $arg (was $savedir)\n";
			$savedir = $arg;
			next;
		}
		else { die "Bad arg for -s\n$usg"; }
	}

	if ($arg eq "-a") {
		$arg = shift @ARGV;
		if (-d $arg) {
			print STDERR "directory = $arg (was $altdir)\n";
			$altdir = $arg;
			next;
		}
		else { die "Bad arg for -a\n$usg"; }
	}

	if ($arg eq "-x") {
		$arg = shift @ARGV;
		print STDERR "exclude = $arg\n";
		$exclude = $arg;
		next;
	}

	if ($arg eq "-ls") {
		$maxhosts = 1;
		$maxtotal = $maxpages = 1000;
		next;
	}

	if ($arg eq "-i") { $img = 1; next; }
	if ($arg eq "-h") { die "$usg"; }
	if ($arg =~ /^-/) { die "Invalid flag\n$usg"; }
	if ($arg =~ /^http:/) {
		if ($start) { die "Please give only one URL\n$usg"; }
		$start = $arg;
	}
	else {
		if ($log) { die "Please give only one output file\n$usg"; }
		$log = $arg;
	}
}

unless ($log) { $log = $deflog; }
$dump = $log;
$dump =~ s/\.[a-z]+/.dump/ || ($dump = "$log.dump");
unless ($start || -f $dump) { die "$usg"; }

(open(STDOUT,">>$log")) || die "Can't create/append $log\n";
$| = 1;
print STDERR "Writing output to $log\n";

if (-f $dump && open (DUMP, $dump)) {
	print STDERR "Reading TODOs from $dump...\n";
	$page = <DUMP>;
	@tocheck = split("\n", $page);
	close (DUMP);
	unlink($dump);
	$start = shift @tocheck;
}

&explore($start);

# Explore the web, starting at $url.
# Maintains a list @tocheck of URLs to try.
sub explore {
	local($url) = @_;
	if ($maxhosts == 1) {
		print "Log of HTML transfer from $url\n$sig\n";
	}
	else {
		print "Log of HTML transfer starting at $url\n";
		if ($maxhosts) {
			print "Maximum hosts to visit = $maxhosts.\n";
		}
		print "Max pages per site = $maxpages.\n",
			"Max total pages = $maxtotal.\n$sig\n";
	}

	$SIG{'INT'} = 'sigint'; # Stop when SIGINT is received
	push(@tocheck,$url);    # Initialize
	$seen{$url} = 1;        # Remember that we've seen it
	($thistype,$thishost, $thisport, $thispath, $request) =
		&url'parse(undef,undef,undef,undef,$url);
	$seenhost{$thishost} = 1;

	while ($#tocheck >= $[) {
		$url = shift @tocheck;
		# Remember current host, port and path:
		($thistype,$thishost, $thisport, $thispath, $request) =
			&url'parse(undef,undef,undef,undef,$url);
		# Check file size:
		$page = &url'http_head($thishost, $thisport, $request);
		$title = split(/\n/, $page, 1);
		if (($title =~ /Server Error:/)
			|| ($title =~ /Bad File Request/)
			|| ($title =~ /404 Not Found/)) {
			print "Invalid page: $url\n";
			next;
		}
		if ($page =~ /Content-length:\s+/) {
			$length = $';
		} else {
			$length = 0;
			print "No content-length: $url\n";
		}
		if (&gotit("$savedir/$request", "$altdir/$request")
		    && &length("$savedir/$request", "$altdir/$request") == $length) {
			# Cached file seems to be valid
			$page = &loadfile("$savedir/$request");
		} else {
			# Not fetched or invalid cache
			print STDERR "Requesting $url\n";
			unless ($page = &url'get($url)) {
				print "Can't get $url\n";
				next;
			} ;

			# Extract the title and detect errors
			if ($page =~ m|<title>([^<]+)</title>|i)
				{ $title = $1; }
			else { $title = "UNTITLED"; }
			if (($title =~ /Server Error:/)
			    || ($title =~ /Bad File Request/)
			    || ($title =~ /404 Not Found/)) {
				print "Invalid page: $url\n";
				next;
			}

			&savefile($savedir . $request, $page);
			&save_url($savedir . $request, $url);
			if (!($maxhosts == 1)) { $hostsig = " ($thishost)"; }
			$title =~ s/\n/ /g;
			# This page is ok, so log it:
			print "$title$hostsig\n";
			# from the last request
			if (++$entries >= $maxtotal) { &dump; last; }
		}
		foreach $href (&hrefs($page)) {
			$href =~ s/#.*$//;
			next if $href eq "";
			# Parse the URL, if possible:
			($type,$host,$port,$path,$request) =
				&url'parse($thistype,$thishost,
					$thisport,$thispath,$href);
			unless ($type eq "http") { next; }
			# Skip this host if invalid:
			unless ($host) { next; }
			# Convert from relative to absolute URL:
			$href = "http://$host:$port$request";
			# Skip if seen already:
			if ($seen{$href}) { next; }
			$seen{$href} = 1;
			# Don't ask too many pages from a given host:
			unless (++$count{$host} <= $maxpages) { next; }
			# Only look at guaranteed .html files:
			unless ($request =~ /\.html$/ || $request =~ /\/$/) {
				if ($exclude && eval "\$request =~ $exclude") {
					next;
				}
			}
			if ($seenhost{$host}) {
				# Seen this host, so add to end of queue:
				print STDERR "Pushing $href\n";
				if ($request =~ /\.html$/
				    || $request =~ /\/$/) {
					# HTML file -- add to checklist
					push(@tocheck,$href);
				} else {
					# Plain file -- add to xfer list
					push (@toget, $href);
				}
			}
			else {
				if ($maxhosts) {
					if (++$hosts > $maxhosts) { next; }
				}
				# New host, so add to front:
				print STDERR "Queueing $href\n";
				if ($request =~ /\.html$/
				    || $request =~ /\/$/) {
					# HTML file -- add to checklist
					unshift(@tocheck,$href);
				} else {
					# Plain file -- add to xfer list
					unshift (@toget, $href);
				}
				$seenhost{$host} = 1;
			}
		}
		if ($img) { push(@toget, &srcs($page)) };
		foreach $url (@toget) {
			# Parse the URL, if possible:
			($type,$host,$port,$path,$request) =
				&url'parse($thistype,$thishost,
					$thisport,$thispath,$url);
			unless ($type eq "http") { next; }
			# Skip this host if invalid:
			unless ($host) { next; }
			# Convert from relative to absolute URL:
			$url = "http://$host:$port$request";
			# Check file size:
			$page = &url'http_head($host, $port, $request);
			$title = split(/\n/, $page, 1);
			if (($title =~ /Server Error:/)
			    || ($title =~ /Bad File Request/)
			    || ($title =~ /404 Not Found/)) {
				print "Invalid IMG: $url\n";
				next;
			}
			if ($page =~ /Content-length:\s+/) {
				$length = $';
				if (&gotit("$savedir/$request", "$altdir/$request")
				    && &length("$savedir/$request", "$altdir/$request") == $length) {
					next;
				}
			} else {
				print "No content-length: $url\n";
			}
			# Skip if seen already:
			if ($seen{$url}) { next; }
			$seen{$url} = 1;

			print STDERR "Requesting $url\n";
			unless ($page = &url'get($url)) {
				print "Can't get $url\n";
				next;
			} ;
			&savefile($savedir . $request, $page);
			&save_url($savedir . $request, $url);
			print "(plain file)\n";
		}
		@toget = ();
	}

	print "\nSearch completed.\n";
	close(STDOUT);
	print STDERR "Result of exploration in $log\n";
}

sub gotit {
	# check if a file has been retrieved already
	local ($path, $alt) = @_;

	return 1 if (-e $path || -e $alt);
	0;
}

sub length {
	local ($path, $alt) = @_;

	return (stat($path))[7] if (-e $path);
	return (stat($alt))[7] if (-e $alt);
}

sub save_url {
	# keep url-to-file mapping information
	local ($path, $url) = @_;
	local ($dir, $file);

	($dir, $file) = &basename($path);
	($url, $file) = &basename($url);
	if (! -e "$dir/.url") {
		open(URL, ">$dir/.url");
		print URL "$url\n";
		close(URL);
	}
}

sub savefile {
	local ($path, $page) = @_;
	local ($dir, $file);

	($dir, $file) = &basename($path);
	# make directories if necessary, and write the page
	if (! -d $dir) {
		if (&mkdirhier($dir)) {
			print "mkdir $dir\n";
		} else {
			print "Error in creating $dir: $!\n";
		}
	}
	if ($path =~ m:/$:) {
		# httpd may have mapped directory to a page,
		# e.g, "/" -> "/foo/Welcome.html".
		# as a workaround, we save it under some name
		# unlikely to conflict.
		$path .= "urlget_dir.html";
	}
	if (open(FILE, ">$path")) {
		print FILE $page;
		close(FILE);
		print "got $path: ";
	} else {
		print "Error in creating $path: $!: ";
	}
}

sub loadfile {
	local ($path) = @_;
	local ($page);
	local ($/);

	if (open(FILE, "$path")) {
		$page = <FILE>;
		close(FILE);
		$page;
	} else {
		print "Error in reading $path: $!\n";
		undef;
	}
}

sub dump {
	# dump @tocheck in a file so that htget can be resumed later.

	open(DUMP, "> $dump");
	print DUMP join("\n", @tocheck);
	close(DUMP);
}

sub sigint {
	local ($/) = "\n";
	local ($res);
	print STDERR "Enter <CR> to continue, \"q\" to quit\n";
	chop ($res = <STDIN>);
	return if ($res eq "");
	if ($res ne "q") {
	    print STDERR "Invalid response -- continuing\n";
	    return;
	}
	print STDERR "Quitting...\n";
	&dump;		# dump @tocheck in a file.
	print "\n\nInterrupted!\n";
	close(STDOUT);
	print STDERR "Result of exploration in $log\n";
	exit(0);
}

# return a list of all the hrefs in a page
sub hrefs {
	local($page) = @_;
	$page =~ s/^[^<]+</</;
	$page =~ s/>[^<]*</></g;
	$page =~ s/>[^<]*$/>/g;
	$page =~ s/<a[^>]*href\s*=\s*"([^"]+)"[^>]*>/$1\n/gi;
	$page =~ s/<[^>]*>//g;
	$page =~ s/\n+/\n/g;
	split(/\n/,$page);
}

# return a list of all the embedded images in a page
sub srcs {
	local($page) = @_;
	$page =~ s/^[^<]+</</;
	$page =~ s/>[^<]*</></g;
	$page =~ s/>[^<]*$/>/g;
	$page =~ s/<img[^>]*src\s*=\s*"([^"]+)"[^>]*>/$1\n/gi;
	$page =~ s/<[^>]*>//g;
	$page =~ s/\n+/\n/g;
	split(/\n/,$page);
}

# Local Variables:
# mode: cperl
# cperl-indent-level: 8
# cperl-continued-statement-offset: 8
# End:
